@@ -1,24 +1,19 @@
 """Common functions."""
 
+from __future__ import annotations
+
 import ast
 import fnmatch
-import imp
 import logging
 import os
-import re
 import sys
 from dataclasses import dataclass, field
+from importlib.util import find_spec
 from pathlib import Path
 from typing import (
     Callable,
-    Dict,
     Generator,
     Iterable,
-    List,
-    Optional,
-    Set,
-    Tuple,
-    Union,
 )
 
 from packaging.markers import Marker
@@ -37,25 +32,25 @@ class FoundModule:
     """A module with uses in the source."""
 
     modname: str
-    filename: str
-    locations: List[Tuple[str, int]] = field(default_factory=list)
+    filename: Path
+    locations: list[tuple[str, int]] = field(default_factory=list)
 
     def __post_init__(self) -> None:
-        self.filename = os.path.realpath(self.filename)
+        self.filename = Path(self.filename).resolve()
 
 
 class _ImportVisitor(ast.NodeVisitor):
     def __init__(self, ignore_modules_function: Callable[[str], bool]) -> None:
         super().__init__()
         self._ignore_modules_function = ignore_modules_function
-        self._modules: Dict[str, FoundModule] = {}
-        self._location: Optional[str] = None
+        self._modules: dict[str, FoundModule] = {}
+        self._location: str | None = None
 
-    def set_location(self, location: str) -> None:
+    def set_location(self, *, location: str) -> None:
         self._location = location
 
     # Ignore the name error as we are overriding the method.
-    def visit_Import(  # pylint: disable=invalid-name
+    def visit_Import(  # noqa: N802, pylint: disable=invalid-name
         self,
         node: ast.Import,
     ) -> None:
@@ -63,7 +58,7 @@ class _ImportVisitor(ast.NodeVisitor):
             self._add_module(alias.name, node.lineno)
 
     # Ignore the name error as we are overriding the method.
-    def visit_ImportFrom(  # pylint: disable=invalid-name
+    def visit_ImportFrom(  # noqa: N802, pylint: disable=invalid-name
         self,
         node: ast.ImportFrom,
     ) -> None:
@@ -79,49 +74,56 @@ class _ImportVisitor(ast.NodeVisitor):
     def _add_module(self, modname: str, lineno: int) -> None:
         if self._ignore_modules_function(modname):
             return
-        path = None
-        progress = []
-        modpath = last_modpath = None
+
+        modname_parts_progress: list[str] = []
         for modname_part in modname.split("."):
+            name = ".".join([*modname_parts_progress, modname_part])
             try:
-                _, modpath, _ = imp.find_module(modname_part, path)
-            except ImportError:
-                # the component specified at this point is not importable
-                # (is just an attribute of the module)
-                # *or* it's not actually installed, so we don't care either
-                break
+                module_spec = find_spec(name=name)
+            except ValueError:
+                # The module has no __spec__ attribute.
+                # For example, if importing __main__.
+                return
 
-            # success! we found *something*
-            progress.append(modname_part)
+            if module_spec is None:
+                # The component specified at this point is not installed.
+                return
 
-            # we might have previously seen a useful path though...
-            if modpath is None:
-                # the `sys` module will hit this code path, and `os` will on
-                # 3.11+.
-                # Possibly others will, but I've not discovered them.
-                modpath = last_modpath
-                break
+            if module_spec.origin is None:
+                modname_parts_progress.append(modname_part)
+                continue
 
-            # ... though it might not be a file, so not interesting to us
-            if not os.path.isdir(modpath):
-                break
+            modpath = module_spec.origin
 
-            path = [modpath]
-            last_modpath = modpath
+            if modpath == "frozen":
+                # Frozen modules are modules written in Python whose compiled
+                # byte-code object is incorporated into a custom-built Python
+                # interpreter by Python's freeze utility.
+                continue
 
-        if modpath is None:
-            # the module doesn't actually appear to exist on disk
+            modpath_path = Path(modpath)
+            modname = module_spec.name
+
+            if modname not in self._modules:
+                if modpath_path.is_file():
+                    if modpath_path.name == "__init__.py":
+                        modpath_path = modpath_path.parent
+                    else:
+                        # We have this empty "else" so that we are
+                        # not tempted to combine the "is file" and "is
+                        # __init__" checks, and to make sure we have coverage
+                        # for this case.
+                        pass
+                self._modules[modname] = FoundModule(
+                    modname=modname,
+                    filename=modpath_path,
+                )
+            assert isinstance(self._location, str)
+            self._modules[modname].locations.append((self._location, lineno))
             return
 
-        modname = ".".join(progress)
-        if modname not in self._modules:
-            self._modules[modname] = FoundModule(modname, modpath)
-        assert isinstance(self._location, str)
-        self._modules[modname].locations.append((self._location, lineno))
-
-    def finalise(self) -> Dict[str, FoundModule]:
-        result = self._modules
-        return result
+    def finalise(self) -> dict[str, FoundModule]:
+        return self._modules
 
 
 def pyfiles(root: Path) -> Generator[Path, None, None]:
@@ -129,41 +131,45 @@ def pyfiles(root: Path) -> Generator[Path, None, None]:
         if root.suffix == ".py":
             yield root.absolute()
         else:
-            raise ValueError(f"{root} is not a python file or directory")
-    elif root.is_dir():
+            msg = f"{root} is not a python file or directory"
+            raise ValueError(msg)
+    else:
         for item in root.rglob("*.py"):
             yield item.absolute()
 
 
 def find_imported_modules(
+    *,
     paths: Iterable[Path],
     ignore_files_function: Callable[[str], bool],
     ignore_modules_function: Callable[[str], bool],
-) -> Dict[str, FoundModule]:
+) -> dict[str, FoundModule]:
     vis = _ImportVisitor(ignore_modules_function=ignore_modules_function)
     for path in paths:
         for filename in pyfiles(path):
             if ignore_files_function(str(filename)):
-                log.info("ignoring: %s", os.path.relpath(filename))
+                log.info("ignoring: %s", filename)
                 continue
-            log.debug("scanning: %s", os.path.relpath(filename))
-            with open(filename, encoding="utf-8") as file_obj:
-                content = file_obj.read()
-            vis.set_location(str(filename))
+            log.debug("scanning: %s", filename)
+            content = filename.read_text(encoding="utf-8")
+            vis.set_location(location=str(filename))
             vis.visit(ast.parse(content, str(filename)))
     return vis.finalise()
 
 
 def find_required_modules(
+    *,
     ignore_requirements_function: Callable[
-        [Union[str, ParsedRequirement]], bool
+        [str | ParsedRequirement],
+        bool,
     ],
     skip_incompatible: bool,
     requirements_filename: Path,
-) -> Set[NormalizedName]:
-    explicit = set()
+) -> set[NormalizedName]:
+    explicit: set[NormalizedName] = set()
     for requirement in parse_requirements(
-        str(requirements_filename), session=PipSession()
+        str(requirements_filename),
+        session=PipSession(),
     ):
         requirement_name = install_req_from_line(
             requirement.requirement,
@@ -176,7 +182,7 @@ def find_required_modules(
 
         if skip_incompatible:
             requirement_string = requirement.requirement
-            if not has_compatible_markers(requirement_string):
+            if not has_compatible_markers(full_requirement=requirement_string):
                 log.debug(
                     "ignoring requirement (incompatible environment "
                     "marker): %s",
@@ -190,7 +196,7 @@ def find_required_modules(
     return explicit
 
 
-def has_compatible_markers(full_requirement: str) -> bool:
+def has_compatible_markers(*, full_requirement: str) -> bool:
     if ";" not in full_requirement:
         return True  # No environment marker.
 
@@ -201,23 +207,29 @@ def has_compatible_markers(full_requirement: str) -> bool:
     return Marker(enviroment_marker).evaluate()
 
 
-def is_package_file(path: str) -> str:
-    """Determines whether the path points to a Python package sentinel
-    file - the __init__.py or its compiled variants.
+def package_path(*, path: Path) -> Path | None:
+    """Return the package path for a given Python package sentinel file.
+
+    Return None if the path is not a sentinel file.
+
+    A sentinel file is the __init__.py or its compiled variants.
     """
-    search_result = re.search(r"(.+)/__init__\.py[co]?$", path)
-    if search_result is not None:
-        return search_result.group(1)
-    return ""
+    if path.parent == path.parent.parent:
+        return None
+
+    if path.name not in ("__init__.py", "__init__.pyc", "__init__.pyo"):
+        return None
+
+    return path.parent
 
 
-def ignorer(ignore_cfg: List[str]) -> Callable[..., bool]:
+def ignorer(*, ignore_cfg: list[str]) -> Callable[..., bool]:
     if not ignore_cfg:
-        return lambda candidate: False
+        return lambda _: False
 
     def ignorer_function(
-        candidate: Union[str, ParsedRequirement],
-        ignore_cfg: List[str] = ignore_cfg,
+        candidate: str | ParsedRequirement,
+        ignore_cfg: list[str] = ignore_cfg,
     ) -> bool:
         for ignore in ignore_cfg:
             if isinstance(candidate, str):
