@@ -1,17 +1,23 @@
 """Find missing requirements."""
 
+from __future__ import annotations
+
 import argparse
 import collections
 import importlib.metadata
 import logging
 import os
 import sys
+from functools import cache
 from pathlib import Path
-from typing import Callable, Iterable, List, Optional, Tuple
+from typing import Callable, Iterable
 from unittest import mock
 
 from packaging.utils import NormalizedName, canonicalize_name
-from pip._internal.commands.show import search_packages_info
+from pip._internal.commands.show import (
+    _PackageInfo,  # pyright: ignore[reportPrivateUsage]
+    search_packages_info,
+)
 from pip._internal.network.session import PipSession
 from pip._internal.req.constructors import install_req_from_line
 from pip._internal.req.req_file import parse_requirements
@@ -22,22 +28,12 @@ from pip_check_reqs.common import FoundModule, version_info
 log = logging.getLogger(__name__)
 
 
-def find_missing_reqs(
-    requirements_filename: Path,
-    paths: Iterable[Path],
-    ignore_files_function: Callable[[str], bool],
-    ignore_modules_function: Callable[[str], bool],
-) -> List[Tuple[NormalizedName, List[FoundModule]]]:
-    # 1. find files used by imports in the code (as best we can without
-    #    executing)
-    used_modules = common.find_imported_modules(
-        paths=paths,
-        ignore_files_function=ignore_files_function,
-        ignore_modules_function=ignore_modules_function,
-    )
-
-    # 2. find which packages provide which files
-    installed_files = {}
+# This is a slow operation.
+# It only happens once when calling the CLI, but it is hit many times in
+# tests.
+# We cache the result to speed up tests.
+@cache
+def get_packages_info() -> list[_PackageInfo]:
     all_pkgs = [
         dist.metadata["Name"] for dist in importlib.metadata.distributions()
     ]
@@ -46,14 +42,32 @@ def find_missing_reqs(
     # dramatically improves speeds.
     # See https://github.com/r1chardj0n3s/pip-check-reqs/issues/123.
     with mock.patch.dict(os.environ, {"_PIP_USE_IMPORTLIB_METADATA": "False"}):
-        packages_info = list(search_packages_info(all_pkgs))
+        return list(search_packages_info(query=all_pkgs))
+
+
+def find_missing_reqs(
+    requirements_filename: Path,
+    paths: Iterable[Path],
+    ignore_files_function: Callable[[str], bool],
+    ignore_modules_function: Callable[[str], bool],
+) -> list[tuple[NormalizedName, list[FoundModule]]]:
+    # 1. find files used by imports in the code (as best we can without
+    #    executing)
+    used_modules = common.find_imported_modules(
+        paths=paths,
+        ignore_files_function=ignore_files_function,
+        ignore_modules_function=ignore_modules_function,
+    )
+
+    installed_files: dict[Path, str] = {}
+    packages_info = get_packages_info()
+    here = Path().resolve()
 
     for package in packages_info:
         package_name = package.name
         package_location = package.location
-        package_files = []
+        package_files: list[str] = []
         for item in package.files or []:
-            here = Path(".").resolve()
             item_location_rel = Path(package_location) / item
             item_location = item_location_rel.resolve()
             try:
@@ -66,14 +80,16 @@ def find_missing_reqs(
             package_files.append(str(relative_item_location))
 
         log.debug(
-            "installed package: %s (at %s)", package_name, package_location
+            "installed package: %s (at %s)",
+            package_name,
+            package_location,
         )
         for package_file in package_files:
-            path = os.path.realpath(
-                os.path.join(package_location, package_file),
-            )
+            path = Path(package_location) / package_file
+            path = path.resolve()
+
             installed_files[path] = package_name
-            package_path = common.is_package_file(path)
+            package_path = common.package_path(path=path)
             if package_path:
                 # we've seen a package file so add the bare package directory
                 # to the installed list as well as we might want to look up
@@ -81,11 +97,14 @@ def find_missing_reqs(
                 installed_files[package_path] = package_name
 
     # 3. match imported modules against those packages
-    used = collections.defaultdict(list)
+    used: collections.defaultdict[
+        NormalizedName,
+        list[common.FoundModule],
+    ] = collections.defaultdict(list)
     for modname, info in used_modules.items():
         # probably standard library if it's not in the files list
         if info.filename in installed_files:
-            used_name = canonicalize_name(installed_files[info.filename])
+            used_name = canonicalize_name(name=installed_files[info.filename])
             log.debug(
                 "used module: %s (from package %s)",
                 modname,
@@ -100,7 +119,7 @@ def find_missing_reqs(
             )
 
     # 4. compare with requirements
-    explicit = set()
+    explicit: set[NormalizedName] = set()
     for requirement in parse_requirements(
         str(requirements_filename),
         session=PipSession(),
@@ -113,11 +132,10 @@ def find_missing_reqs(
         log.debug("found requirement: %s", requirement_name)
         explicit.add(canonicalize_name(requirement_name))
 
-    result = [(name, used[name]) for name in used if name not in explicit]
-    return result
+    return [(name, used[name]) for name in used if name not in explicit]
 
 
-def main(arguments: Optional[List[str]] = None) -> None:
+def main(arguments: list[str] | None = None) -> None:
     usage = "usage: %prog [options] files or directories"
     parser = argparse.ArgumentParser(usage)
     parser.add_argument("paths", type=Path, nargs="*")
@@ -174,14 +192,14 @@ def main(arguments: Optional[List[str]] = None) -> None:
     parse_result = parser.parse_args(arguments)
 
     if parse_result.version:
-        print(version_info())
+        sys.stdout.write(version_info() + "\n")
         sys.exit(0)
 
     if not parse_result.paths:
         parser.error("no source files or directories specified")
 
-    ignore_files = common.ignorer(parse_result.ignore_files)
-    ignore_mods = common.ignorer(parse_result.ignore_mods)
+    ignore_files = common.ignorer(ignore_cfg=parse_result.ignore_files)
+    ignore_mods = common.ignorer(ignore_cfg=parse_result.ignore_mods)
 
     logging.basicConfig(format="%(message)s")
     if parse_result.debug:
@@ -209,7 +227,7 @@ def main(arguments: Optional[List[str]] = None) -> None:
             for filename, lineno in use.locations:
                 log.warning(
                     "%s:%s dist=%s module=%s",
-                    os.path.relpath(filename),
+                    filename,
                     lineno,
                     name,
                     use.modname,
